#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include "vga.h"
#include <vgamouse.h>

#define	SOLID_LINE	0xFFFF
#define DOTTED_LINE	0xAAAA
#define CENTER_LINE	0xF2F2
#define DASHED_LINE	0xC3C3
#define OPEN		0
#define FILLED		1
#define FILL		1

#define MOUSE		1
#define NOMOUSE		0

#ifndef MAXPATTERN
#define MAXPATTERN	64
#endif

char *font, *paintings;
FILE *fontfile;
fd_set rfds;
struct timeval tv;
static struct termios stored_settings;
int X__min, X__max=639, Y__min, Y__max=479;   	/* graphical window  */
int X__org, Y__org;				/* position of (0,0) */
int __linestyle=0xFFFF;				/* line style        */

void init_keyboard(void)
        {
        struct termios new_settings;
        tcgetattr(0,&stored_settings);
        new_settings = stored_settings;
        /* Disable canonical mode, and set buffer size to 1 byte */
        new_settings.c_lflag &= (~ICANON);
        new_settings.c_cc[VTIME] = 0;
        new_settings.c_cc[VMIN] = 1;
        /* Don't echo characters on the screen */
/*        new_settings.c_lflag &= (~ECHO); */
        tcsetattr(0,TCSANOW,&new_settings);
        return;
        }

void restore_keyboard(void)
        {
        tcsetattr(0,TCSANOW,&stored_settings);
        return;
        }

int kbhit(void)
        {
        /* Watch stdin (fd 0) to see when it has input. */
        FD_ZERO(&rfds);
        FD_SET(0, &rfds);
        tv.tv_sec = 0;
        tv.tv_usec = 1;
        return select(1, &rfds, NULL, NULL, &tv);
        }

void plotpix(int x, int y)
	{
	if(x>=X__min && x<=X__max && y>=Y__min && y<=Y__max)
		{
		if(__linestyle & 1)
			{
			vga_drawpixel(x+X__org, y+Y__org);
			__linestyle >>= 1;
			__linestyle |= 0x8000;
			}
		else
			__linestyle >>= 1;
		}
	}

int getpix(int x, int y)
        {
        if(x>=X__min && x<=X__max && y>=Y__min && y<=Y__max)
		return vga_getpixel(x+X__org, y+Y__org);
	else
		return -1;
	}

void drawline(int x1, int y1, int x2, int y2)
	{
	int len_x, len_y, errorterm, x_sign, y_sign;
	int pix_x, pix_y;
	pix_x=x1;
	pix_y=y1;
	x_sign=x2-x1;
	len_x=abs(x_sign);
	y_sign=y2-y1;
	len_y=abs(y_sign);
	if(len_y>=len_x)
		{
		errorterm=(-1*len_y)/2;
		while(pix_y!=y2)
			{
			plotpix(pix_x, pix_y);
			/* drawl_upd_y */
			errorterm += len_x;
			if(y_sign<0) pix_y--;
			else pix_y++;
			if(errorterm>=0)
				{
				/* drawl_upd_x */
				errorterm -= len_y;
				if(x_sign<0) pix_x--;
				else pix_x++;
				}
			}
		}
	else
		{
		errorterm=len_x/2;
		while(pix_x!=x2)
			{
			plotpix(pix_x, pix_y);
			/* drawl_upd_x */
                        errorterm -= len_y;
                        if(x_sign<0) pix_x--;
                        else pix_x++;
			if(errorterm<0) 
			/* drawl_upd_y */
				{
			        errorterm += len_x;
       		                if(y_sign<0) pix_y--;
                        	else pix_y++;
				}
			}
		}
	}

void block(int mode, int xlo, int ylo, int xrb, int yrb)
	{
	int x,y;
	if(mode)
		for(x=xlo; x<=xrb; x++)
			for(y=yrb; y<=ylo; y++)
				plotpix(x,y);	
	else
		{
	        drawline(xlo, ylo, xlo, yrb);
        	drawline(xlo, yrb, xrb, yrb);
        	drawline(xrb, yrb, xrb, ylo);
        	drawline(xrb, ylo, xlo, ylo);
		}
	}

void circle(int mode, int p, int q, int r)
	{
	int x,y,s,i;
	x=0;
	y=r;
	s=1-2*r;
	while(x<=y)
		{
		if(!mode)
			{
			plotpix(p+x, q+y);
			plotpix(p-x, q-y);
			plotpix(p+y, q-x);
			plotpix(p-y, q+x);
			if(x!=0 && x!=y)
				{
				plotpix(p+y, q+x);
				plotpix(p-x, q+y);
				plotpix(p+x, q-y);
				plotpix(p-y, q-x);
				}
			}
		else
			{
			for(i=p-x; i<=p+x; i++)
				{
				plotpix(i,q+y);
				plotpix(i,q-y);
				}
			for(i=p-y; i<=p+y; i++)
				{
				plotpix(i, q+x);
				plotpix(i, q-x);
				}
			}
		s=s+4*x+2;
		x++;
		if(s>0)
			{
			s=s-4*y+4;
			y--;
			}
		}
	}

void gwrite(char *text, int x, int y, int hsize, int vsize)
	{
	int row, col, byte, p=0, h, v, t, line;
	while(text[p])
		{
		row=y;
		for(line=0; line<16; line++)
			{
			byte = font[text[p]*16+line];
			for(v=0; v<vsize; v++)
				{
				col=x+p*hsize*8;
				t=0x80;
				while(t != 0)
					{
					if(byte & t)
						for(h=0; h<hsize; h++)
							vga_drawpixel(col+h,row+v);
					t = t/2;
					col += hsize;
					}
				}
			row += vsize;
			}
		p++;
		}
	}

int pow(int base, int n)
        {
        int p;
        for(p=1; n>0; --n)
                p=p*base;
        return p;
        }

void printdec(int v, int n, int x, int y, int hsize, int vsize)
	{
	int digit, p=0;
	char *s;
	s=malloc(2);
	if (s==NULL)
		{
		printf("Cannot allocate 2 bytes for decimal printing :-(\n");
		restore_keyboard();
        	vga_setmode(TEXT);
		exit(1);
		}
	v %= pow(10,n);
	while(n>0)
		{
		s[1]=0x00;
		digit = v/pow(10, n-1);
		if (digit || p || n==1)
			{
			s[0]='0' + digit;
			gwrite(s, x, y, hsize, vsize);
			p=1;
			}
		v %= pow(10, n-1);
		x += 24;
		n--;
		}
	free(s);
	}

void scroll(char *s, int scrollen, char c, int x, int y, int hsize, int vsize, int colour, int bgcol)
	{
	char *msg;
	int k=0, i, n, p=0;
	msg=malloc(scrollen);
	if(msg==NULL)
		{
		printf("Could not allocate 80 bytes for scrolling message\n");
		restore_keyboard();
	        vga_setmode(TEXT);
		exit(1);
		}
	if(kbhit()) k=getchar();
	while(tolower(k)!=c)
		{
		n=p;
		for(i=0; i<scrollen; i++)
			{
			msg[i]=s[n++];
			if(s[n]=='\0') n=0;
			}
		msg[scrollen-1]='\0';
		p++;
		if(s[p]=='\0') p=0;
		vga_setcolor(colour);
		gwrite(msg, x, y, hsize, vsize);
		usleep(250000);
		vga_setcolor(bgcol);
                gwrite(msg, x, y, hsize, vsize);
		if(kbhit()) k=getchar();
		}
	free(msg);
	}

int loadfont(char *f)
	{
	int i;
	if(font==NULL)
		font=malloc(4096);
	if(font)
		if ((fontfile = fopen(f, "rb")) != NULL)
			{
			for(i=0; i<4096; i++)
				font[i] = fgetc(fontfile);
			fclose(fontfile);
			return 0;	/* Font loaded */
			}
		else
			return 1;	/* Could not open font file */
	else return 2;			/* Memory allocation failed */
	}

void grmod(int mouz)
	{
	X__min = Y__min = X__org = Y__org = 0;
	X__max=639;
	Y__max=479;
	__linestyle=0xFFFF;
	vga_init();
	vga_setmousesupport(mouz);
	vga_setmode(4);
	vga_clear();
	}

void txmod(void)
	{
	vga_setmode(TEXT);
	}

void gwindow(int xlo, int ylo, int xrb, int yrb)
	{
	X__min=xlo+X__org;
	X__max=xrb+X__org;
	Y__min=yrb+Y__org;
	Y__max=ylo+Y__org;
	}

void gwinoff(void)
	{
        X__min=-X__org;
	Y__min=-Y__org;
        X__max=649-X__org;
        Y__max=479-Y__org;
	}

void origin(int x_org, int y_org)
        {
        X__org=x_org;
        Y__org=y_org;
        gwinoff();
        }

void setlinestyle(int style)
	{
	__linestyle=style;
	}

/* Paint functies:
-	paint_def_pattern	definieert patroon
-	paint 			kleurt in, patroonnr of -1 voor huidige kleur
*/

int paintdef(int nr, char *pattern)
	{
	int i;
        if(paintings==NULL)
                paintings=malloc(256*MAXPATTERN);
        if(paintings)
		if(nr < MAXPATTERN)
                        {
                        for(i=0; i<256; i++)
                                paintings[i+256*nr] = pattern[i];
                        return 0;	/* No errors */
			}
		else return 1;		/* Bad pattern number */
        else return 2;			/* Memory allocation failed */
	}

int paint(int x, int y, int border, int nr)
	{
	int xl, xr, xp, yp, pat;
	if(paintings || nr<0)
		if(nr < MAXPATTERN)
			{
			pat=256*nr;
			xp=x;
			yp=y;
		        while(getpix(xp, yp)!=border && yp>=Y__min)
				{
				xr=xp;
				xl=xp-1;
				while(getpix(xr, yp)!=border && xr<=X__max)
					{
					if(nr>=0) vga_setcolor(paintings[pat+xr%16+16*(yp%16)]);
					plotpix(xr++, yp);
					}
		        	while(getpix(xl, yp)!=border && xl>=X__min)
	        	        	{
                                        if(nr>=0) vga_setcolor(paintings[pat+xl%16+16*(yp%16)]);
		        	        plotpix(xl--, yp);
					}
				xp=xl+(xr-xl)/2;
				yp--;
				}
			yp=y+1;
		        xp=x;
			while(getpix(xp, yp)!=border && yp<=Y__max)
				{
		                xr=xp;
		                xl=xp-1;
	        	        while(getpix(xr, yp)!=border && xr<=X__max)
	                	        {
                                        if(nr>=0) vga_setcolor(paintings[pat+xr%16+16*(yp%16)]);
	                        	plotpix(xr++, yp);
		                        }	
		                while(getpix(xl, yp)!=border && xl>=X__min)
	        	                {
                                        if(nr>=0) vga_setcolor(paintings[pat+xl%16+16*(yp%16)]);
	                	        plotpix(xl--, yp);
	                        	}
				xp=xl+(xr-xl)/2;
		                yp++;
				}
			return 0;	/* Everything ok */
			}
		else	return 1;	/* Bad pattern number */
	else	return 2;		/* No patterns defined */
	}	

void triangle(int mode, int px1, int py1, int px2, int py2, int px3, int py3)
	{
/*	int i, xlinks[Y__max+1], xrechts[Y__max+1];*/
	int i, xlinks[640], xrechts[640];
        int len_x, len_y, errorterm, x_sign, y_sign;
        int pix_x, pix_y, x1=0, y1=0, x2=0, y2=0;
	if(mode)
		{
		for(i=0; i<=Y__max; i++)
			{
			xlinks[i]=X__max+1;
			xrechts[i]=X__min-1;
			}
		for(i=1; i<4; i++)
			{
			switch(i)
				{
				case 1:
					{
					x1=px1;
					y1=py1;
					x2=px2;
					y2=py2;
					break;
					}
				case 2:
					{
					x1=px2;
					y1=py2;
					x2=px3;
					y2=py3;
					break;
					}
				case 3:
					{
					x1=px1;
					y1=py1;
					x2=px3;
					y2=py3;
					break;
					}
				}
		       	pix_x=x1;
	       		pix_y=y1;
	        	x_sign=x2-x1;
		        len_x=abs(x_sign);
		        y_sign=y2-y1;
	        	len_y=abs(y_sign);
	       	 	if(len_y>=len_x)
		                {
	        	        errorterm=(-1*len_y)/2;
	                	while(pix_y!=y2)
		                        {
					if(pix_x<=xlinks[pix_y] && pix_y<=Y__max && pix_y>=Y__min) xlinks[pix_y]=pix_x;
					if(pix_x>=xrechts[pix_y] && pix_y<=Y__max && pix_y>=Y__min) xrechts[pix_y]=pix_x;
	        	                /* drawl_upd_y */
	                	        errorterm += len_x;
		                        if(y_sign<0) pix_y--;
		                        else pix_y++;
	        	                if(errorterm>=0)
	                	                {
	                        	        /* drawl_upd_x */
	                                	errorterm -= len_y;
		                                if(x_sign<0) pix_x--;
		                                else pix_x++;
	        	                        }
	                	        }
	       	        	 }	
		        else	
	       		        {
	               		errorterm=len_x/2;
	                	while(pix_x!=x2)
	        	                {
	                                if(pix_x<=xlinks[pix_y] && pix_y<=Y__max && pix_y>=Y__min) xlinks[pix_y]=pix_x;
	                                if(pix_x>=xrechts[pix_y] && pix_y<=Y__max && pix_y>=Y__min) xrechts[pix_y]=pix_x;
	               	        	/* drawl_upd_x */
	                	        errorterm -= len_y;
	       	                	if(x_sign<0) pix_x--;
	                	        else pix_x++;
	       	                	if(errorterm<0)
	        	                /* drawl_upd_y */
	       	        	                {
	               	        	        errorterm += len_x;
	                       	        	if(y_sign<0) pix_y--;
	                                	else pix_y++;
		                                }
	       		                }
	               		}
	                        if(pix_x<=xlinks[pix_y] && pix_y<=Y__max && pix_y>=Y__min) xlinks[pix_y]=pix_x;
	                        if(pix_x>=xrechts[pix_y && pix_y<=Y__max] && pix_y>=Y__min) xrechts[pix_y]=pix_x;
			}
		if(py1<=py2 && py1<=py3) y1=py1;
		if(py2<=py1 && py2<=py3) y1=py2;
		if(py3<=py1 && py3<=py2) y1=py3;
	        if(py1>=py2 && py1>=py3) y2=py1;
	        if(py2>=py1 && py2>=py3) y2=py2;
	        if(py3>=py1 && py3>=py2) y2=py3;
		for(i=y1; i<=y2; i++)
			drawline(xlinks[i], i, xrechts[i], i);
		}
	else
		{
		drawline(px1, py1, px2, py2);
		drawline(px2, py2, px3, py3);
		drawline(px3, py3, px1, py1);
		}
	}

int gread(char *buffer, int x, int y, int hsize, int vsize, int n, int col, int bgcol)
	{
	int xl, i=0, c=0;
	char *input;
	char letter[]="*";
	
	vga_setcolor(col);
	input=malloc(n+1);
	xl=x;
	while(c!='\n')
		{
		c=getchar();
		if(c==127 || c==8)
			{
			if(i>0)
				{
				letter[0]=input[i-1];
				vga_setcolor(bgcol);
				gwrite(letter, xl+(i-1)*hsize*8, y, hsize, vsize);
				vga_setcolor(col);
				i--;
				input[i]=0x00;
				}
			}
		else
			{
			if(i<n && c>31)
				{
				input[i]=c;
				input[i+1]=0x00;
				letter[0]=c;
				gwrite(letter, xl+i*hsize*8, y, hsize, vsize);
				x+=hsize;
				i++;
				}
			}
		}
	input[i]=0x00;
	strcpy(buffer,input);
	free(input);
	return(i-1);
	}
