#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include "vga.h"

int i, x, y, timer, minlengte, maxkleur, maxdoel, maxtijd;
int done=0, score, level, tijd, goal1, goal2, goal3, goal4, toets;
int g1tab[]={5,10,5,5,10}, g2tab[]={10,5,5,5,10}, g3tab[]={5,5,10,5,10}, g4tab[]={5,5,5,10,10};
char *font, colormap[15], veld[100], rij[10];
FILE *fontfile;
fd_set rfds;
struct timeval tv;
static struct termios stored_settings;

#define	maskgoal1 0x80
#define maskgoal2 0x40
#define maskgoal3 0x20
#define maskgoal4 0x10

void set_keypress(void)
        {
        struct termios new_settings;
        tcgetattr(0,&stored_settings);
        new_settings = stored_settings;
        /* Disable canonical mode, and set buffer size to 1 byte */
        new_settings.c_lflag &= (~ICANON);
        new_settings.c_cc[VTIME] = 0;
        new_settings.c_cc[VMIN] = 1;
        /* Don't echo characters on the screen */
/*        new_settings.c_lflag &= (~ECHO); */
        tcsetattr(0,TCSANOW,&new_settings);
        return;
        }

void reset_keypress(void)
        {
        tcsetattr(0,TCSANOW,&stored_settings);
        return;
        }

int kbhit(void)
        {
        /* Watch stdin (fd 0) to see when it has input. */
        FD_ZERO(&rfds);
        FD_SET(0, &rfds);
        tv.tv_sec = 0;
        tv.tv_usec = 1;
        return select(1, &rfds, NULL, NULL, &tv);
        }

void block(int xlo, int ylo, int xrb, int yrb, int c)
	{
	int n;
	vga_setcolor(colormap[c&0x0F]);
	for(n=xlo; n<=xrb; n++)
		vga_drawline(n,ylo,n,yrb);
	}

void markeer(int x, int y)
	{
	int xlo, ylo, xrb, yrb;
	xlo=x*36+20;
	ylo=(10-y)*34+37;
	xrb=x*36+52;
	yrb=(10-y)*34+69;
	vga_setcolor(colormap[15]);
	vga_drawline(xlo, ylo, xlo, yrb);
	vga_drawline(xlo, yrb, xrb, yrb);
	vga_drawline(xrb, yrb, xrb, ylo);
	vga_drawline(xrb, ylo, xlo, ylo);
	}
	
void gwrite(char *text, int x, int y, int c, int hsize, int vsize)
	{
	int row, col, byte, p=0, h, v, t, line;
	vga_setcolor(colormap[c]);
	while(text[p])
		{
		row=y;
		for(line=0; line<16; line++)
			{
			byte = font[text[p]*16+line];
			for(v=0; v<vsize; v++)
				{
				col=x+p*hsize*8;
				t=0x80;
				while(t != 0)
					{
					if(byte & t)
						for(h=0; h<hsize; h++)
							vga_drawpixel(col+h,row+v);
					t = t/2;
					col += hsize;
					}
				}
			row += vsize;
			}
		p++;
		}
	}

void zetveld(int x, int y, int v)
	{
	if(y == 10)
		rij[x]=v;
	else
		veld[x*10+y]=v;
	}

int leesveld(int x, int y)
	{
	if(y==10)
		return rij[x];
	else
		return veld[x*10+y];
	}

void steen(int x, int y, int c)
	{
	if (y==10)
		block(x*36+20, 50, x*36+52, 18, c);
	else
		block(x*36+20, (10-y)*34+37, x*36+52, (10-y)*34+69, c);
	}

void val(void)
	{
	int x,y;

	for(x=0; x<10; x++)
		for(y=9; y>-1; y--)
			if(leesveld(x,y)==0)
				{
				steen(x, y, leesveld(x, y+1));
				zetveld(x, y, leesveld(x, y+1));
				steen(x, y+1, 0); 
				zetveld(x, y+1, 0);
				}
	}

void schuiflinks(void)
	{
	int x, s;
	s = rij[0];
	for(x=0; x<9; x++)
		{
		rij[x]=rij[x+1];
		steen(x, 10, rij[x]);
		}
	rij[9] = s;
	steen(9, 10, s);
	}

void schuifrechts(void)
	{
	int x, s;
	s = rij[9];
	for(x=9; x>0; x--)
		{
		rij[x] = rij[x-1];
		steen(x, 10, rij[x]);
		}
	rij[0] = s;
	steen(0, 10, s);
	}

int pow(int base, int n)
        {
        int p;
        for(p=1; n>0; --n)
                p=p*base;
        return p;
        }

void printdec(int v, int x, int y, int n)
	{
	int digit, p=0;
	char *s;
	s=malloc(2);
	if (s==NULL)
		{
		printf("Cannot allocate 2 bytes for decimal printing :-(\n");
		reset_keypress();
        	vga_setmode(TEXT);
		exit(1);
		}
	v %= pow(10,n);
	while(n>0)
		{
		s[0]=0x01;
		s[1]=0x00;
		gwrite(s, x, y, 0, 3, 2);
		digit = v/pow(10, n-1);
		if (digit || p || n==1)
			{
			s[0]='0' + digit;
			gwrite(s, x, y, 3, 3, 2);
			p=1;
			}
		v %= pow(10, n-1);
		x += 24;
		n--;
		}
	free(s);
	}

void zoek(void)
	{
	int x, y, cel, t, bingo=1;
	int xw, yw;
	while(bingo)
		{
		bingo=0;
		for(x=0; x<10; x++)
			for(y=0; y<10; y++)
				{
				cel=leesveld(x,y);
				if(cel)
					{
					if(!(cel & maskgoal1))
						{
						cel &= 0x0F;
						t=0;
						xw=x;
						while(xw<10 && (leesveld(xw, y) & 0x0F) == cel)
							{
							xw++;
							t++;
							}
						if(t >= minlengte)
							{
							bingo++;
							xw=x;
							while(t>0)
								{
								markeer(xw, y);
								zetveld(xw, y, leesveld(xw, y) | maskgoal1);
								t--;
								xw++;
								}
							if (goal1)
								{
								goal1--;
								printdec(goal1, 411, 303, 2);
								}
							}
						}
					if(!(cel & maskgoal2))
                                                {
                                                cel &= 0x0F;
                                                t=0;
                                                yw=y;
                                                while(yw<10 && (leesveld(x, yw) & 0x0F) == cel)
                                                        {
                                                        yw++;
                                                        t++;
                                                        }
                                                if(t >= minlengte)
                                                        {
                                                        bingo++;
                                                        yw=y;
                                                        while(t>0)
                                                                {
                                                                markeer(x, yw);
                                                                zetveld(x, yw, leesveld(x, yw) | maskgoal2);
                                                                t--;
                                                                yw++;
                                                                }
							if (goal2) 
								{
								goal2--;
								printdec(goal2, 526, 303, 2);
								}
                                                        }
                                                }
					if(!(cel & maskgoal3))
                                                {
                                                cel &= 0x0F;
                                                t=0;
						xw=x;
                                                yw=y;
                                                while(xw<10 && yw<10 && (leesveld(xw, yw) & 0x0F) == cel)
                                                        {
							xw++;
                                                        yw++;
                                                        t++;
                                                        }
                                                if(t >= minlengte)
                                                        {
                                                        bingo++;
							xw=x;
                                                        yw=y;
                                                        while(t>0)
                                                                {
                                                                markeer(xw, yw);
                                                                zetveld(xw, yw, leesveld(xw, yw) | maskgoal3);
                                                                t--;
								xw++;
                                                                yw++;
                                                                }
							if (goal3) 
								{
								goal3--;
								printdec(goal3, 411, 370, 2);
								}
                                                        }
                                                }
					if(!(cel & maskgoal4))
                                                {
                                                cel &= 0x0F;
                                                t=0;
                                                xw=x;
                                                yw=y;
                                                while(xw<10 && yw>-1 && (leesveld(xw, yw) & 0x0F) == cel)
                                                        {
                                                        xw++;
                                                        yw--;
                                                        t++;
                                                        }
                                                if(t >= minlengte)
                                                        {
                                                        bingo++;
                                                        xw=x;
                                                        yw=y;
                                                        while(t>0)
                                                                {
                                                                markeer(xw, yw);
                                                                zetveld(xw, yw, leesveld(xw, yw) | maskgoal4);
                                                                t--;
                                                                xw++;
                                                                yw--;
                                                                }
							if (goal4) 
								{
								goal4--;
								printdec(goal4, 526, 370, 2);
								}
                                                        }
                                                }
					}
				}
		for(x=0; x<10; x++)
			for(y=0; y<10; y++)
				while(leesveld(x, y) & 0xF0)
					{
					t=y+1;
					while (t<10)
						{
						cel=leesveld(x, t);
						zetveld(x, t-1, cel);
						steen(x, t-1, cel);
						t++;
						}
					zetveld(x, 9, 0);
					steen(x, 9, 0);
					score++;
					printdec(score, 426, 91, 7);
					}
		}
	}
			
void printmsg(char *s)
	{
	char *msg;
	int k=0, i, n, p=0;
	msg=malloc(80);
	if(msg==NULL)
		{
		printf("Could not allocate 80 bytes for scrolling message\n");
		reset_keypress();
	        vga_setmode(TEXT);
		exit(1);
		}
        block(20,461,619,429, 0);
	if(kbhit()) k=getchar();
	while(tolower(k)!='s')
		{
		n=p;
		for(i=0; i<73; i++)
			{
			msg[i]=s[n++];
			if(s[n]=='\0') n=0;
			}
		msg[73]='\0';
		p++;
		if(s[p]=='\0') p=0;
		gwrite(msg, 25, 437, 15, 1, 1);
		usleep(250000);
                gwrite(msg, 25, 437, 0, 1, 1);
		if(kbhit()) k=getchar();
		}
	block(20,461,619,429, 0);
	free(msg);
	}

void main(void)
	{
	int key;
	/* Initialize keyboard (Disable canonical mode, and set buffer size to 1 byte, no echo on screen */
	set_keypress();

	/* Allocate font memory */
	font = malloc(4096);
	if (font == NULL)
		{
		printf("Allocation of font memory (4096 bytes) failed.\n");
	        reset_keypress();
	        vga_setmode(TEXT);
		exit(1);
		} 
	/* Load font data */
	if ((fontfile = fopen("/usr/lib/games/triples.fnt", "rb")) != NULL)
		{
		for(i=0; i<4096; i++)
			font[i] = fgetc(fontfile);
		fclose(fontfile);
		}
	else
		{
		printf("Couldn't open font `triples.fnt'\n");
	        reset_keypress();
        	vga_setmode(TEXT);
		exit(1);
		}


	/* Initialise graphics screen and build screen */
	vga_init();
	vga_setmode(4); 
	vga_clear();
	for(i=0; i<16; i++)
		colormap[i] = i;
	colormap[3]  =  4;			/* red */
	colormap[4]  = 14;			/* yellow */
	colormap[6]  =  3;			/* cyan */
	colormap[14] =  6;			/* brown */

	block(0,479,639,0, 14);			/* background */
	block(20,409,52,71, 0);			/* column 1 */
	block(56,409,88,71, 0);               	/* column 2 */
	block(92,409,124,71, 0);              	/* column 3 */
	block(128,409,160,71, 0);             	/* column 4 */
	block(164,409,196,71, 0);             	/* column 5 */
	block(200,409,232,71, 0);             	/* column 6 */
	block(236,409,268,71, 0);             	/* column 7 */
	block(272,409,304,71, 0);             	/* column 8 */
	block(308,409,340,71, 0);             	/* column 9 */
	block(344,409,376,71, 0);             	/* column 10 */

	block(400,131,619,71, 0);		/* score display */
	block(400,233,500,173, 0);		/* level display */
	block(520,233,619,173, 0);		/* time display */
	block(400,409,619,275, 0);		/* goals display */
	block(20,461,619,429, 0);		/* message display */

	/* print "Triples" logo and display titles */
	gwrite("TRIPLES", 414, 21, 0, 3, 2);
	gwrite("TRIPLES", 411, 18, 15, 3, 2);
	gwrite("score", 416, 75, 15, 1, 1);
	gwrite("level", 408, 177, 15, 1, 1);
	gwrite("time", 528, 177, 15, 1, 1);
	gwrite("goal", 408, 279, 15, 1, 1);
	gwrite("-", 447,287, 2, 1, 1);
	gwrite("|", 557,287, 2, 1, 1);
	gwrite("/", 447, 338, 2, 1, 1);
	gwrite("\\", 557, 338, 2, 1, 1);

	while(!done)
		{
		printmsg("Triples for Linux by Roland Leurs *** Press 's' to start *** Keys: "
			 "key < : move top row left, key > : move the top row right, "
			 "space bar: drop the top row down, q : quit the game *** ");
		for(x=0; x<10; x++)
			for(y=0; y<11; y++)
				{
				zetveld(x,y,0);
				steen(x,y,0);
				}
		level=0;					/* reset level */
		score=0;					/* reset score */
		maxtijd=90;					/* reset playing time */
		while(maxtijd && !done)				/* first loop, if maxtijd=0 then you're the champion */
			{
			minlengte = 3;
			while(minlengte < 11 && !done)		/* second loop, if minlengte>10 then play it faster */
				{
				maxkleur = 3;
				while(maxkleur < 11 && !done)	/* third loop, if maxkleur>10 then enlarge the chain */
					{
					maxdoel = 0;
					while(maxdoel < 5 && !done)	/* fourth loop, if maxdoel > 4 then add a color */
						{
						tijd = maxtijd;
						level += 1;
						goal1 = g1tab[maxdoel];
						goal2 = g2tab[maxdoel];
						goal3 = g3tab[maxdoel];
						goal4 = g4tab[maxdoel];
						printdec(score, 426, 91, 7);
						printdec(goal1, 411, 303, 2);
                                                printdec(goal2, 526, 303, 2);
                                                printdec(goal3, 411, 370, 2);
                                                printdec(goal4, 526, 370, 2);
						printdec(level, 410, 200, 3);
						printdec(tijd, 530, 200, 2);
						timer=20;
						block(20,461,619,429, 0);
						while((goal1 || goal2 || goal3 || goal4) && !done)
							{
							done=1;					/* playground full ? */
                                                	for(x=0; x<10; x++)
                                                        	{
								if(!leesveld(x,9)) done=0; 	/* no, still playing */
                                                        	if(rij[x] == 0)
                                                        	        {
                                                        	        rij[x]=1 + random() % maxkleur;
                                                                	steen(x,10,rij[x]);
                                                                	}
                                                        	}
							if(kbhit()) key=getchar();
							else key=0;
							if (tolower(key)=='q') done=2;    /* player quits */
							if (key==',') schuiflinks();
							if (key=='.') schuifrechts();
							if (key==' ')
								{
								val();
								zoek();
								}
							usleep(50000);
							if(--timer == 0)
								{
								tijd--;
								printdec(tijd, 530, 200, 2);
								timer=20;
								}
							if(tijd<15)
								{
								if (timer==20) 
									gwrite("Warning: you're running out of time", 41, 437, 14, 2, 1);
								if (timer==10)
									gwrite("Warning: you're running out of time", 41,437, 0, 2, 1);
								}
							if(tijd<0) done=4; /* time out */
							} 
						while(tijd>0 && !done)
							{
							score += 2;
							printdec(score, 426, 91, 7);
							tijd -= 1;
							printdec(tijd, 530, 200, 2);
							}
						maxdoel += 1;
						}	
					maxkleur += 1;
					}
				minlengte += 1;
				}
			maxtijd -= 45;
			}
		if(done!=2) done=0;
	}
	free(font);					
	/* Restore system */
	reset_keypress();
	vga_setmode(TEXT);
	exit(0);
	}
